# coding: utf-8
u"""Базовые классы для асинхронных задач."""
from __future__ import absolute_import

from collections import OrderedDict
from datetime import datetime
import logging

from celery import Task
from celery import states

from . import models
from . import statuses


class AsyncTask(Task):

    u"""Базовый класс асинхронных задач."""

    abstract = True

    # описание (имя) асинх.задачи (отображается в реестре)
    description = None

    def debug(self, *args, **kwargs):
        u"""логирование в debug-лог."""
        logger = logging.getLogger('educommon.async')
        logger.debug(*args, **kwargs)

    def apply_async(self, args=None, kwargs=None, task_id=None, producer=None,
                    link=None, link_error=None, **options):
        u"""Постановка задачи в очередь.

        автор задачи задаётся снаружи через 2 поля в словаре kwargs:
            object_id и content_type
        """
        if kwargs is None:
            kwargs = {}

        async_result = super(AsyncTask, self).apply_async(
            args=args, kwargs=kwargs, task_id=task_id, producer=producer,
            link=link, link_error=link_error, **options
        )

        location = self.__class__.__module__ + '.' + self.__class__.__name__
        task_meta = models.AsyncTaskMeta.objects.create(
            location=location,
            description=self.description
        )
        params = dict(
            task_meta=task_meta,
            status=statuses.STATUS_RECEIVED,
            object_id=kwargs.get('object_id'),
            content_type=kwargs.get('content_type'),
            task_type_id=kwargs.get(
                'task_type', models.AsyncTaskType.TASK_UNKNOWN
            ),
            # время начала задачи
            # для случая, когда celery успевает начать задачу ещё до
            # коммита в БД создаваемой здесь models.RunningTask
            queued_on=datetime.now()
        )

        models.RunningTask.objects.get_or_create(
            task_id=async_result.task_id, defaults=params
        )

        self.debug("Task %s added" % self.__name__)

        return async_result

    def run(self, *args, **kwargs):
        u"""Выполнение задачи (отложенное)."""
        # начальное состояние задачи
        self.state = {
            # результаты задачи
            'values': OrderedDict(),
            # описание результата
            'description': self.description,
            # прогресс выполнения задачи
            'progress': u'Неизвестно'
        }
        self.update_state(
            state=statuses.get_state_str(statuses.STATUS_STARTED),
            meta=self.state
        )
        self.debug("Task %s run (task_id = %s )" % (
            self.__name__, self.request.id))
        return {}

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        u"""Завершение задачи."""
        self.debug("Task %s completed" % self.__name__)

        if isinstance(retval, dict):
            self.update_state(
                state=retval.get('task_state', status), meta=retval)
        else:
            self.update_state(state=status, meta=retval)

    def update_state(self, task_id=None, state=None, meta=None):
        u"""Обновленение модели при изменении состояния асинх. результата.

        :param str task_id: id задачи celery
        :param str task_state: состояние задачи celery (celery.states)
        :param meta: состояние метаданых (см. базовый класс в Celery).
        """
        if task_id is None:
            task_id = self.request.id

        running_task = models.RunningTask.objects.filter(
            task_id=task_id
        ).order_by('queued_on').first()

        if not running_task:
            return

        if state == states.SUCCESS:
            self.state['values'][u'Время выполения'] = (
                datetime.now() - running_task.queued_on)

        super(AsyncTask, self).update_state(
            task_id=task_id, state=state, meta=meta)

        running_task.status = statuses.get_status_idx(state)

        if state == states.STARTED:
            # если обновление статуса связано с переходом в состояние 'RUNNING'
            running_task.queued_on = datetime.now()
        running_task.clean_and_save()

    def set_progress(self, task_id=None, task_state=states.STARTED,
                     progress=None, values=None):
        u"""Обновление состояния выполнения задачи.

        :param str task_id: id задачи celery
        :param str task_state: состояние задачи celery (celery.states)
        :param unicode progress: строковое описание состояния выполнения задачи
        :param dict values: значения задаваемые процедурой выполнения задачи
        """
        if task_id is None:
            task_id = self.request.id

        if progress:
            self.state['progress'] = progress
        if values:
            self.state['values'].update(values)
        self.backend.store_result(task_id, self.state, task_state)
