# coding: utf-8
u"""Валидаторы для простых (текстовых, числовых и т.п.) полей модели Django."""
from __future__ import absolute_import

from functools import partial
from itertools import cycle
import re

from django.core.validators import DecimalValidator
from django.core.exceptions import ValidationError
from django.core.exceptions import NON_FIELD_ERRORS
from django.utils.deconstruct import deconstructible
from six.moves import range
from six.moves import zip
import magic
import six

from educommon.utils.misc import get_mime_type_for_extension

from . import IModelValidator
from . import validate_value
from ..utils import get_original_object


# =============================================================================
# СНИЛС
# =============================================================================
_snils_re = re.compile(r'^\d{3}-\d{3}-\d{3} \d{2}$')  # СНИЛС


def snils_validator(value):
    u"""Валидатор СНИЛС.

    Проверяет корректность формата и контрольное число (последние две цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату СНИЛС, либо контрольное
        число некорректно.
    """
    value = six.text_type(value)

    if not _snils_re.match(value):
        raise ValidationError(u'СНИЛС должен быть в формате ###-###-### ##')

    if value[:11] <= u'001-001-998':
        return

    numbers = (int(ch) for ch in reversed(value[:11]) if ch.isdigit())
    checksum = int(value[12:14])
    summa = sum(i * n for i, n in enumerate(numbers, 1))
    if summa > 101:
        summa %= 101

    if (summa < 100 and summa != checksum or
            summa in (100, 101) and checksum != 0):
        raise ValidationError(u'не пройдена проверка контрольного числа')


is_snils_valid = partial(validate_value, validator=snils_validator)


# =============================================================================
# ИНН
# =============================================================================


def _check_inn_checksum(coefficients, numbers, checksum):
    summa = sum(c * n for c, n in zip(coefficients, numbers))
    if summa % 11 % 10 != checksum:
        raise ValidationError(u'Не пройдена проверка контрольного числа')


def inn10_validator(value):
    u"""Валидатор для ИНН юридического лица (10 цифр).

    Проверяет корректность формата и контрольное число (последняя цифра).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = six.text_type(value)

    if len(value) != 10 or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 10-тизначным числом')

    _check_inn_checksum((2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-1]),
                        int(value[-1]))


def inn12_validator(value):
    u"""Валидатор для ИНН физического лица (12 цифр).

    Проверяет корректность формата и контрольное число (последние 2 цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = six.text_type(value)

    if len(value) != 12 or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 12-тизначным числом')

    _check_inn_checksum((7, 2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-2]),
                        int(value[-2]))

    _check_inn_checksum((3, 7, 2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-1]),
                        int(value[-1]))


def inn_validator(value):
    u"""Валидатор для поля ИНН.

    Проверяет корректность формата и контрольное число (1 или 2 последние
    цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = six.text_type(value)

    if len(value) not in (10, 12) or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 10-ти или 12-тизначным числом')

    if len(value) == 10:
        inn10_validator(value)
    else:
        inn12_validator(value)


is_inn_valid = partial(validate_value, validator=inn_validator)
is_inn10_valid = partial(validate_value, validator=inn10_validator)
is_inn12_valid = partial(validate_value, validator=inn12_validator)


# =============================================================================
# КПП, ОКАТО, ОКТМО, ОКПО, ОГРН, ОКВЭД, ОКОПФ, ОКФС
# =============================================================================


def kpp_validator(value):
    u"""Валидатор для КПП (кода причины постановки на налоговый учет)."""
    value = six.text_type(value)

    if len(value) != 9 or not value.isdigit():
        raise ValidationError(u'КПП должен быть 9-тизначным числом')


is_kpp_valid = partial(validate_value, validator=kpp_validator)


def checksum_pr_50_1_024(value):
    u"""Расчет контрольного числа в соответствии с ПР 50.1.024-2005.

    Контрольное числорассчитывается следующим образом:
        1. Разрядам кода в общероссийском классификаторе, начиная со старшего
           разряда,присваивается набор весов, соответствующий натуральному
           ряду чисел от 1 до 10. Если разрядность кода больше 10, то набор
           весов повторяется.
        2. Каждая цифра кода умножается на вес разряда и вычисляется сумма
           полученных произведений.
        3. Контрольное число для кода представляет собой остаток от деления
           полученной суммы на модуль 11.
        4. Контрольное число должно иметь один разряд, значение которого
           находится в пределах от 0 до 9.

    Если получается остаток, равный 10, то для обеспечения одноразрядного
    контрольного числа необходимо провести повторный расчет, применяя вторую
    последовательность весов, сдвинутую на два разряда влево (3, 4, 5,...).

    Если в случае повторного расчета остаток от деления вновь сохраняется
    равным 10, то значение контрольного числа проставляется равным 0.
    """
    summa = sum(
        int(digit) * weight
        for digit, weight in zip(value, cycle(list(range(1, 11))))
    )

    remainder = summa % 11
    if remainder < 10:
        result = remainder
    else:
        summa = sum(
            int(digit) * weight
            for digit, weight in zip(value, cycle(list(range(3, 13))))
        )
        remainder = summa % 11
        if remainder < 10:
            result = remainder
        else:
            result = 0

    return result


def okato_validator(value):
    u"""Валидатор для ОКАТО.

    ОКАТО - общероссийский классификатор объектов административно-
    территориального деления.

    .. seealso::
       `Контрольное число ОКАТО <http://kontragent.info/articles/okato>`_
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКАТО должен состоять только из цифр')

    value_length = len(value)
    if value_length in (3, 6, 9, 12):
        # В последнем разряде указано контрольное число, проверим его
        if not checksum_pr_50_1_024(value[:-1]) != int(value[-1]):
            raise ValidationError(u'Не пройдена проверка контрольного числа')

    elif value_length not in (2, 5, 8, 11):
        raise ValidationError(
            u'Код ОКАТО должен состоять из 2, 5, 8 или 11 цифр'
        )


is_okato_valid = partial(validate_value, validator=okato_validator)


def oktmo_validator(value):
    u"""Валидатор для ОКТМО.

    ОКТМО - Общероссийский классификатор территорий муниципальных образований.
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКТМО должен состоять только из цифр')

    value_length = len(value)
    if value_length not in (2, 5, 8, 11):
        raise ValidationError(
            u'ОКТМО не может состоять из {} цифр'.format(value_length)
        )


is_oktmo_valid = partial(validate_value, validator=oktmo_validator)


def okpo_validator(value):
    u"""Валидатор кода ОКПО.

    ОКПО - Общероссийский классификатор предприятий и организаций.
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКПО должен состоять только из цифр')

    value_length = len(value)
    if value_length not in (8, 10):
        raise ValidationError(u'ОКПО должен состоять из 8-ми или 10-ти цифр')


is_okpo_valid = partial(validate_value, validator=okpo_validator)


def _check_ogrn_checksum(value):
    u"""Проверка контрольного числа ОГРН(ИП).

    Младший разряд остатка от деления предыдущего (n-1)-значного числа на (n-2)
    должен быть равен n-ому знаку ОГРН, где n – длина ОГРН. Если остаток от
    деления равен 10, то контрольное число равно 0 (нулю).
    """
    value_length = len(value)  # количество цифр в ОГРН
    divisor = value_length - 2  # делитель для вычисления контрольного числа
    number = int(value[:value_length - 1])  # ОГРН
    control_number = int(value[-1])  # контрольное число

    if number % divisor % 10 != control_number:
        raise ValidationError(u'Не пройдена проверка контрольного числа')


def _ogrn_validator(value, valid_length, title):
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОГРН должен состоять только из цифр')
    if len(value) not in valid_length:
        raise ValidationError(
            u'{} должен состоять из {} цифр!'.format(
                title,
                u' или '.join(six.text_type(l) for l in valid_length)
            )
        )

    _check_ogrn_checksum(value)


def ogrn13_validator(value):
    u"""Валитор ОГРН.

    ОГРН - Основной государственный регистрационный номер.
    """
    _ogrn_validator(value, (13,), u'ОГРН')


def ogrn15_validator(value):
    u"""Валидатор ОГРНИП.

    ОГРНИП - Основной государственный регистрационный номер индивидуального
    предпринимателя.
    """
    _ogrn_validator(value, (15,), u'ОГРНИП')


def ogrn_validator(value):
    u"""Валидатор ОГРН(ИП).

    ОГРН(ИП) - Основной государственный регистрационный номер (индивидуального
    предпринимателя).
    """
    _ogrn_validator(value, (13, 15), u'ОГРН(ИП)')


is_ogrn_valid = partial(validate_value, validator=ogrn_validator)
is_ogrn13_valid = partial(validate_value, validator=ogrn13_validator)
is_ogrn15_valid = partial(validate_value, validator=ogrn15_validator)


okved_re = re.compile(r'^(\d{2}|\d{2}\.\d{1,2}|\d{2}\.\d{1,2}\.\d{1,2})$')


def okved_validator(value):
    u"""Валидатор кода ОКВЭД.

    ОКВЭД - Общероссийский классификатор видов экономической деятельности.
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКВЭД должен состоять только из цифр')

    value_length = len(value)
    if value_length < 2 or 6 < value_length:
        raise ValidationError(
            u'Код ОКВЭД не может состоять из {} цифр'.format(value_length)
        )


is_okved_valid = partial(validate_value, validator=okved_validator)


okopf_re = re.compile(r'^(\d \d\d \d\d)$')


def okopf_validator(value):
    u"""Валидатор кода ОКОПФ.

    ОКОПФ - Общероссийский классификатор организационно-правовых форм.
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКОПФ должен состоять только из цифр')

    value_length = len(value)
    if value_length != 5:
        raise ValidationError(u'Код ОКОПФ должен состоять из 5-ти цифр')


is_okopf_valid = partial(validate_value, validator=okopf_validator)


def okfs_validator(value):
    u"""Валидатор кода ОКФС.

    ОКФС - Общероссийский классификатор форм собственности.
    """
    value = six.text_type(value)

    if not value.isdigit():
        raise ValidationError(u'ОКФС должен состоять только из цифр')

    value_length = len(value)
    if value_length != 2:
        raise ValidationError(u'Код ОКФС должен состоять из 2-ти цифр')


is_okfs_valid = partial(validate_value, validator=okfs_validator)


def validate_file_mime_type(file, allowed_mime_types):
    u"""Валидация mimetype файла.

    При помощи библиотеки python-magic получает mimetype файла и проверять, что
    данный тип в ходит в allowe_mime_types.

    :param file: Проверяемый файл
    :type file: file
    :param allowed_mime_types: Набор Mime types допустимых для загрузки
    :type: iterable
    """
    file_mime_type = magic.from_buffer(file.read(), mime=True)
    if file_mime_type not in allowed_mime_types:
        raise ValidationError(
            u'Mime type файла не соотвествует допустимым в системе'
        )


@deconstructible
class FileMimeTypeValidator(object):

    """Валидатор для FileField модели проверяющий mimetype файла."""

    message = (
        'Mime type файла не допустим для загрузки в системе.'
    )
    code = 'invalid_mimetype'

    def __init__(self, allowed_extensions, message=None, code=None):
        self.allowed_extensions = [
            allowed_extension.lower()
            for allowed_extension
            in allowed_extensions
        ]

        self.allowed_mime_types = set(
            get_mime_type_for_extension(extension)
            for extension in self.allowed_extensions
            if get_mime_type_for_extension(extension)
        )

        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        try:
            validate_file_mime_type(value, self.allowed_mime_types)
        except ValidationError:
            raise ValidationError(
                self.message,
                code=self.code,
            )


# =============================================================================
# Общие валидаторы на основе классов
# =============================================================================
class OptionalFieldValidator(IModelValidator):
    u"""
    Валидатор поля, которое являетсяся обязательным при выполнении условий.

    Пример:

    .. code-block::

        class OptionalCommentValidator(OptionalFieldValidator):
            # Если статус равен OTHER, должен быть указан комментарий
            def  _need_check(self, instance):
                return instance.status == Status.OTHER

        class Model(BaseModel):
            status = # ...
            status_comment = models.TextField(null=True, blank=True, ...)
            validators = [OptionalCommentValidator('status_comment'),]
    """

    def __init__(self, field_name, **kwargs):
        u"""Инициализация валидатора.

        :param str field_name: имя валидируемого поля
        """
        assert isinstance(field_name, str), type(field_name)
        self.field_name = field_name
        self.empty_values = (None, '', [], (), {})

    def _need_check(self, instance):
        u"""Проверяет, выполнены ли условия для проверки поля ``field_name``.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :rtype: bool
        """
        raise NotImplementedError()

    def _get_message(self, instance):
        u"""Возвращает сообщение об ошибке.

        Может быть переопределено в реализациях валидатора для вывода
        более подробных сообщений.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :rtype: str
        """
        return u'Обязательное поле.'

    def clean(self, instance, errors):
        u"""Проверяет заполненность поля.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :param errors: ошибки, выявленные в ходе проверки.
        :type errors: collections.OrderedDict
        """
        if self._need_check(instance):
            if getattr(instance, self.field_name, None) in self.empty_values:
                errors[self.field_name].append(self._get_message(instance))


class RequiredBooleanValidator(OptionalFieldValidator):

    u"""Валидатор обязательного поля ``BooleanField``.

    Пример:

    .. code-block:: python

        class Model(BaseModel):
            accept_rules = models.BooleanField(...)
            validators = [RequiredBooleanValidator('accept_rules'),]
    """

    def __init__(self, field_name, **kwargs):
        u"""Инициализация валидатора.

        :param str field_name: имя валидируемого поля
        """
        super(RequiredBooleanValidator, self).__init__(field_name, **kwargs)
        self.empty_values = (False,)

    def _need_check(self, instance):
        u"""
        Проверяет, выполнены ли условия для проверки поля ``field_name``.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :rtype bool
        """
        return True


class RequiredFieldValidator(OptionalFieldValidator):
    u"""Валидатор поля, не обязательного на уровне базы данных.

    Пример:

    .. code-block:: python

        class Model(BaseModel):
            # Поле не обязательно в основной системе, но в рамках плагина
            # должно быть обязательным
            employee = models.ForeignKey(
                'employee.Employee',
                null=True, blank=True,
            )
            validators = [RequiredFieldValidator('employee'),]
    """

    def _need_check(self, instance):
        u"""
        Проверяет, выполнены ли условия для проверки поля ``field_name``.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :rtype bool
        """
        return True


class UnchangeableFieldValidator(IModelValidator):

    u"""Валидатор неизменяемого поля.

    При редактировании объекта, если изменяется поле ``field_name``,
    возвращает ошибку.

    Пример:

    .. code-block:: python

        class Model(BaseModel):
            type = models.BooleanField(...)
            validators = [UnchangeableFieldValidator('type'),]

    """

    def __init__(self, field_name, **kwargs):
        u"""Инициализация валидатора.

        :param str field_name: имя валидируемого поля
        """
        assert isinstance(field_name, str), type(field_name)
        self.field_name = field_name

    def _get_message(self, instance):
        u"""Возвращает сообщение об ошибке.

        Может быть переопределено в реализациях валидатора для вывода
        более подробных сообщений.

        :param instance: экземпляр проверяемой модели.
        :type instance: django.db.models.base.Model

        :rtype: str
        """
        verbose_name = instance._meta.get_field(
            self.field_name
        ).verbose_name
        return u'Поле "{}" не доступно для редактирования.'.format(
            verbose_name
        )

    def clean(self, instance, errors):
        u"""Проверка редактирования поля ``field_name``.

        :param instance: экземпляр проверяемой модели.
        :type instance: .mixins.validation.ModelValidationMixin

        :param errors: ошибки, выявленные в ходе проверки.
        :type errors: collections.OrderedDict
        """
        original = get_original_object(instance)
        if not original:
            return

        old_value = getattr(original, self.field_name, None)
        new_value = getattr(instance, self.field_name, None)
        if old_value != new_value:
            errors[self.field_name].append(self._get_message(instance))


class DuplicationValidator(IModelValidator):

    u"""Валидатор уникальности объекта.

    Используется для объектов, уникальность которых нельзя проверить
    через ``Meta.unique_together``.

    Пример:

    .. code-block:: python

        class MyModelDuplicationValidator(DuplicationValidator):
            # Проверяет по полям type и status (если оно задано)

            def _get_field_names(self, instance):
                result = ['Тип',]
                if instance.status:
                    result.append('Статус')
                return result

            def _get_duplication_params(self, instance):
                query = Q(type=instance.type)
                if instance.status:
                    query &= Q(status=instance.status)
                return query

        class Model(BaseModel):
            type = # ...
            status = # ...
            validators = [MyModelDuplicationValidator(),]
    """

    def _get_duplication_params(self, instance):
        u"""Возвращает параметры для проверки уникальности.

        :param instance: экземпляр проверяемой модели.
        :type instance: django.db.models.base.Model

        :rtype django.db.models.Q
        """
        raise NotImplementedError()

    def _get_field_names(self, instance):
        u"""Возвращает список названий полей, по которым происходит проверка.

        Получает валидируемый объект, т.к. проверяемый список полей может
        зависеть от значений полей объекта.

        :param instance: экземпляр проверяемой модели.

        :return список строк-названий полей
        :rtype list
        """
        raise NotImplementedError()

    def _get_message(self, instance):
        u"""Возвращает сообщение об ошибке.

        Может быть переопределено в реализациях валидатора для вывода
        более подробных сообщений.

        :param instance: экземпляр проверяемой модели.
        :type instance: django.db.models.base.Model

        :rtype: str
        """
        return (
            u'В системе уже существует {} с такими же значениями полей {}.'
        ).format(
            instance._meta.verbose_name,
            ', '.join(self._get_field_names(instance))
        )

    def _get_base_query(self, instance):
        u"""
        Возвращает базовый кварисет, по которому проверяется уникальность.

        :param instance: экземпляр проверяемой модели.
        :type instance: django.db.models.base.Model

        :rtype: django.db.models.QuerySet
        """
        return instance._meta.model.objects

    def clean(self, instance, errors):
        u"""Проверка уникальности объекта.

        :param instance: экземпляр проверяемой модели.
        :type instance: django.db.models.base.Model

        :param errors: ошибки, выявленные в ходе проверки.
        :type errors: collections.OrderedDict
        """
        original = get_original_object(instance)
        query = self._get_base_query(instance).filter(
            self._get_duplication_params(instance)
        )

        if original:
            query = query.exclude(id=instance.id)

        if query.exists():
            errors[NON_FIELD_ERRORS].append(self._get_message(instance))


class SingleErrorDecimalValidator(DecimalValidator):

    u"""Кастомный класс валидации Decimal поля модели."""

    def __init__(self, max_digits, decimal_places):
        super(SingleErrorDecimalValidator, self).__init__(max_digits, decimal_places)
        self.max_whole_digits = max_digits - decimal_places

    def __call__(self, value):
        u"""
        Переопределенный стандартный метод валилидации Decimal поля.

        При привышенеии целой и дробной части выводит сообщение такое же как и
        при привышении только целой части. Код ошибки остался max_digits.
        """
        digit_tuple, exponent = value.as_tuple()[1:]
        decimals = abs(exponent)
        digits = len(digit_tuple)
        if decimals > digits:
            digits = decimals
        whole_digits = digits - decimals

        if self.max_digits is not None and digits > self.max_digits:
            raise ValidationError(
                self.messages['max_whole_digits'],
                code='max_digits',
                params={'max': self.max_whole_digits},
            )
        if self.decimal_places is not None and decimals > self.decimal_places:
            raise ValidationError(
                self.messages['max_decimal_places'],
                code='max_decimal_places',
                params={'max': self.decimal_places},
            )
        if (self.max_digits is not None and self.decimal_places is not None and
                whole_digits > (self.max_digits - self.decimal_places)):
            raise ValidationError(
                self.messages['max_whole_digits'],
                code='max_whole_digits',
                params={'max': self.max_whole_digits},
            )
