# coding: utf-8
from __future__ import absolute_import

from m3.db import safe_delete
from m3_django_compat import atomic
from m3_django_compat import get_model

from educommon.utils.db import get_related_instances


class CascadeDeletePackMixin(object):

    u"""Mixin определяющий исключаемые поля модели при проверке DeleteCheck.

    Поскольку DeleteCheck срабатывает до вызова delete_row возникают ситуации,
    когда нам требуется при работе с этим паком исключить определенные
    поля модели, из которых имеются ссылки на модель. Для решения данных
    ситуаций используется атрибут cascade_delete_objects.

    .. code-block:: python

        cascade_delete_objects = (
            ( 'parent', 'parentportfolio', 'person'),
        )
    """
    # содержит кортежи формата
    # (имя приложения, имя модели приложения, имя поля)
    cascade_delete_objects = ()

    def _cascade_delete(self, obj_id, request, context):
        for app_label, model_name, field_name in (
                self.__class__.cascade_delete_objects):
            model = get_model(app_label, model_name)
            lookup = {
                field_name + '_id': obj_id
            }
            for obj in model.objects.filter(**lookup):
                if hasattr(obj, 'safe_delete') and callable(obj.safe_delete):
                    obj.safe_delete()
                else:
                    safe_delete(obj)

    @classmethod
    def skip_field(cls, field):
        if cls.cascade_delete_objects:
            for app_label, model_name, field_name in (
                    cls.cascade_delete_objects):
                model = get_model(app_label, model_name)
                if field.name == field_name and field.model == model:
                    return True
        return False

    @atomic
    def delete_row(self, obj_id, request, context):
        self._cascade_delete(obj_id, request, context)
        return super(CascadeDeletePackMixin, self).delete_row(
            obj_id, request, context)


class CascadeDeleteMixin(object):

    u"""Определяет связи, исключаемые при проверке в ``DeleteCheck``.

    Перед удалением объекта модели обеспечивает удаление зависимых от него
    объектов, если настройки зависимых объектов (``cascade_delete_for``)
    позволяют это делать.

    .. code-block:: python
       :caption: Пример использования

       class ModelA(CascadeDeleteMixin, models.Model):

           field = models.TextField()

       class ModelB(models.Model):

           link = models.ForeignKey(ModelA)
           cascade_delete_for = (link,)

    При удалении объектов этой модели с помощью метода ``safe_delete()`` также
    будут удалены объекты модели ``ModelB``, которые ссылаются на ``ModelA``
    через поле ``link``. Это произойдет потому, что в
    ``ModelB.cascade_delete_for`` указано это поле.
    """

    @staticmethod
    def skip_field(field):
        return field in getattr(field.model, 'cascade_delete_for', ())

    @atomic
    def safe_delete(self):
        for obj in get_related_instances(
            self,
            collapse=False,
            skip_func=lambda field: not self.skip_field(field),
        ):
            if hasattr(obj, 'safe_delete') and callable(obj.safe_delete):
                obj.safe_delete()
            else:
                safe_delete(obj)

        return super(CascadeDeleteMixin, self).safe_delete()
